@@ -34,6 +34,16 @@ def Indent2(s):
   return '\n'.join('  ' + l for l in s.split('\n'))
 
 
+LogicalVariable = collections.namedtuple(
+  'LogicalVariable',
+  [
+    'variable_name',    # Name of user or generated variable.
+    'predicate_name',   # Name of a predicate in rule for which the variable is
+                        # used.
+    'is_user_variable'  # Whether this is a user variable (vs generated one).
+  ])
+
+
 class RuleCompileException(Exception):
   """Exception thrown when user-error is detected at rule-compile time."""
 
@@ -67,14 +77,18 @@ def HeadToSelect(head):
     else:
       assert 'expression' in v, 'Bad select value: %s' % str(v)
       select[k] = v['expression']  # <=> v as k
+  # Allowing predicates with no arguments.
+  if not select:
+    select['atom'] = {'literal': {'the_string': {'the_string': 'yes'}}}
 
   return (select, aggregated_vars)
 
 
-def AllMentionedVariables(x, dive_in_combines=False):
+def AllMentionedVariables(x, dive_in_combines=False, this_is_select=False):
   """Extracting all variables mentioned in an expression."""
   r = []
-  if isinstance(x, dict) and 'variable' in x:
+  # In select there can be a variable named variable.
+  if isinstance(x, dict) and 'variable' in x and not this_is_select:
     r.append(x['variable']['var_name'])
   if isinstance(x, list):
     for v in x:
@@ -182,6 +196,7 @@ class RuleStructure(object):
     self.tables = collections.OrderedDict()
     # Table variable to clause variable map.
     self.vars_map = {}
+    self.vars_heritage_map = {}
     # Clause variable to one table variable.
     self.inv_vars_map = {}
     self.vars_unification = []
@@ -193,9 +208,23 @@ class RuleStructure(object):
     self.allocator = names_allocator
     self.external_vocabulary = external_vocabulary
     self.synonym_log = {}
-    self.full_rull_text = None
+    self.full_rule_text = None
     self.distinct_denoted = None
 
+  def SelectAsRecord(self):
+    def StrIntKey(x):
+      k, v = x
+      if isinstance(k, str):
+        return (k, v)
+      if isinstance(k, int):
+        return ('%03d' % k, v)
+      assert False, 'x:%s' % str(x)    
+    return {'record': {
+      'field_value': [{
+        'field': k,
+        'value': {'expression': v}
+      } for k, v in sorted(self.select.items(), key=StrIntKey)]}}
+
   def OwnVarsVocabulary(self):
     """Returns a map: logica variable -> SQL expression with the value."""
     def TableAndFieldToSql(table, field):
@@ -225,7 +254,7 @@ class RuleStructure(object):
 
   def AllVariables(self):
     r = set()
-    r |= AllMentionedVariables(self.select)
+    r |= AllMentionedVariables(self.select, this_is_select=True)
     r |= AllMentionedVariables(self.vars_unification)
     r |= AllMentionedVariables(self.constraints)
     r |= AllMentionedVariables(self.unnestings)
@@ -258,12 +287,32 @@ class RuleStructure(object):
             self.full_rule_text)
     self.unnestings = ordered_unnestings
 
-  def ElliminateInternalVariables(self, assert_full_ellimination=False):
+  def ReplaceVariableEverywhere(self, u_left, u_right):
+    if 'variable' in u_right:
+      l = self.synonym_log.get(u_right['variable']['var_name'], [])
+      l.append(LogicalVariable(variable_name=u_left,
+                               predicate_name=self.this_predicate_name,
+                               # TODO: sqlite_recursion somehow gets
+                               # u_left to be int. 
+                               is_user_variable=(isinstance(u_left, str) and
+                                                 not u_left.startswith('x_'))))
+      l.extend(self.synonym_log.get(u_left, []))
+      self.synonym_log[u_right['variable']['var_name']] = l
+    ReplaceVariable(u_left, u_right, self.unnestings)
+    ReplaceVariable(u_left, u_right, self.select)
+    ReplaceVariable(u_left, u_right, self.vars_unification)
+    ReplaceVariable(u_left, u_right, self.constraints)
+    
+  # TODO: Parameter unfold_recods just patches some bug. Careful review is needed.
+  def ElliminateInternalVariables(self, assert_full_ellimination=False, unfold_records=True):
     """Elliminates internal variables via substitution."""
     variables = self.InternalVariables()
     while True:
       done = True
+      self.vars_unification = [
+          u for u in self.vars_unification if u['left'] != u['right']]
       for u in self.vars_unification:
+        # Direct variable assignments.
         for k, r in [['left', 'right'], ['right', 'left']]:
           if u[k] == u[r]:
             continue
@@ -279,16 +328,47 @@ class RuleStructure(object):
                   not str(u[k]['variable']['var_name']).startswith('x_'))):
             u_left = u[k]['variable']['var_name']
             u_right = u[r]
-            if 'variable' in u_right:
-              l = self.synonym_log.get(u_right['variable']['var_name'], [])
-              l.append(u_left)
-              l.extend(self.synonym_log.get(u_left, []))
-              self.synonym_log[u_right['variable']['var_name']] = l
-            ReplaceVariable(u_left, u_right, self.unnestings)
-            ReplaceVariable(u_left, u_right, self.select)
-            ReplaceVariable(u_left, u_right, self.vars_unification)
-            ReplaceVariable(u_left, u_right, self.constraints)
+            self.ReplaceVariableEverywhere(u_left, u_right)
             done = False
+        # Assignments to variables in record fields.
+        if unfold_records:  # Confirm that unwraping works and make this unconditional.
+          # Unwrapping goes wild sometimes. Letting it go right to left only.
+          # for k, r in [['left', 'right']]:
+          for k, r in [['left', 'right'], ['right', 'left']]:
+            if u[k] == u[r]:
+              continue
+            ur_variables = AllMentionedVariables(u[r])
+            ur_variables_incl_combines = AllMentionedVariables(
+                u[r], dive_in_combines=True)
+            if (isinstance(u[k], dict) and
+                'record' in u[k] and
+                ur_variables <= self.ExtractedVariables()):
+              def AssignToRecord(target, source):
+                global done
+                for fv in target['record']['field_value']:
+                  def MakeNewSource():
+                    return {
+                      'subscript': {
+                        'record': source,
+                        'subscript': {'literal': {'the_symbol': {'symbol': fv['field']}}}
+                      }
+                    }
+                  if ('variable' in fv['value']['expression'] and
+                      fv['value']['expression']['variable']['var_name']
+                        in variables and
+                      fv['value']['expression']['variable']['var_name']
+                        not in ur_variables_incl_combines):
+                    u_left = fv['value']['expression']['variable']['var_name']
+                    u_right = MakeNewSource()
+                    self.ReplaceVariableEverywhere(u_left, u_right)
+                    done = False
+                  if 'record' in fv['value']['expression']:
+                    new_target = fv['value']['expression']
+                    new_source = MakeNewSource()
+                    AssignToRecord(new_target, new_source)
+
+              AssignToRecord(u[k], u[r])
+      
       if done:
         variables = self.InternalVariables()
         if assert_full_ellimination:
@@ -296,26 +376,46 @@ class RuleStructure(object):
             if variables:
               violators = []
               for v in variables:
-                violators.extend(self.synonym_log.get(v, []))
+                violators.extend(
+                  v.variable_name 
+                  for v in self.synonym_log.get(v, [])
+                  if v.predicate_name == self.this_predicate_name)
                 violators.append(v)
               violators = {v for v in violators if not v.startswith('x_')}
-              assert violators, (
-                  'Logica needs better error messages: purely internal '
-                  'variable was not eliminated. It looks like you have '
-                  'not passed a required argument to some called predicate. '
-                  'Use --add_debug_info_to_var_names flag to make this message '
-                  'a little more informatvie. '
-                  'Variables: %s, synonym_log: %s' % (str(variables),
-                                                      str(self.synonym_log)))
               # Remove disambiguation suffixes from variables not to confuse
               # the user.
-              violators = {v.split(' # disambiguated')[0] for v in violators}
-              raise RuleCompileException(
-                  color.Format(
-                      'Found no way to assign variables: '
-                      '{warning}{violators}{end}. '
-                      'This error might also come from injected sub-rules.',
-                      dict(violators=', '.join(sorted(violators)))),
+              if violators:
+                violators = {v.split(' # disambiguated')[0] for v in violators}
+                raise RuleCompileException(
+                    color.Format(
+                        'Found no way to assign variables: '
+                        '{warning}{violators}{end}.',
+                        dict(violators=', '.join(sorted(violators)))),
+                    self.full_rule_text)
+              else:
+                user_variables = [
+                  uv for v in variables 
+                  for uv in self.synonym_log.get(v, [])
+                  if uv.is_user_variable]
+                this_predicate = color.Format('{warning}{p}{end}',
+                                              dict(p=self.this_predicate_name))
+                unassigned_vars = ', '.join(
+                  color.Format('{warning}{var}{end} in rule for '
+                               '{warning}{p}{end}',
+                               dict(var=v.variable_name, p=v.predicate_name))
+                  for v in user_variables
+                )
+                assert user_variables, (
+                    'Logica needs better error messages: purely internal '
+                    'variable was not eliminated. It looks like you have '
+                    'not passed a required argument to some called predicate. '
+                    'Use --add_debug_info_to_var_names flag to make this message '
+                    'a little more informatvie. '
+                    'Variables: %s, synonym_log: %s' % (str(variables),
+                                                        str(self.synonym_log)))
+                raise RuleCompileException(
+                  'While compiling predicate ' + this_predicate + ' there was '
+                  'found no way to assign variables: ' + unassigned_vars + '.',
                   self.full_rule_text)
           else:
             assert not variables, (
@@ -400,7 +500,11 @@ class RuleStructure(object):
 
     for k, v in self.select.items():
       if k == '*':
-        fields.append('%s.*' % ql.ConvertToSql(v))
+        if 'variable' in v:
+          v['variable']['dont_expand'] = True  # For SQLite.
+        fields.append(
+          subquery_encoder.execution.dialect.Subscript(
+           ql.ConvertToSql(v), '*', True))
       else:
         fields.append('%s AS %s' % (ql.ConvertToSql(v), LogicaFieldToSqlField(k)))
     r += ',\n'.join('  ' + f for f in fields)
@@ -428,21 +532,28 @@ class RuleStructure(object):
           tables.append(sql)
       self.SortUnnestings()
       for element, the_list in self.unnestings:
+        if 'variable' in element:
+          # To prevent SQLite record unfolding.
+          element['variable']['dont_expand'] = True
         tables.append(
             subquery_encoder.execution.dialect.UnnestPhrase().format(
                 ql.ConvertToSql(the_list), ql.ConvertToSql(element)))
       if not tables:
-        tables.append('(SELECT "singleton" as s) as unused_singleton')
+        tables.append("(SELECT 'singleton' as s) as unused_singleton")
       from_str = ', '.join(tables)
       # Indent the from_str.
       from_str = '\n'.join('  ' + l for l in from_str.split('\n'))
       r += from_str
       if self.constraints:
-        r += '\nWHERE\n'
         constraints = []
+        # Predicates used for type inference.
+        ephemeral_predicates = ['~']
         for c in self.constraints:
-          constraints.append(ql.ConvertToSql(c))
-        r += ' AND\n'.join(map(Indent2, constraints))
+          if c['call']['predicate_name'] not in ephemeral_predicates:
+            constraints.append(ql.ConvertToSql(c))
+        if constraints:
+          r += '\nWHERE\n'
+          r += ' AND\n'.join(map(Indent2, constraints))
       if self.distinct_vars:
         ordered_distinct_vars = [
             v for v in self.select.keys() if v in self.distinct_vars]
@@ -455,7 +566,8 @@ class RuleStructure(object):
                          for v in ordered_distinct_vars)
         elif subquery_encoder.execution.dialect.GroupBySpecBy() == 'expr':
           r += ', '.join(
-            ql.ConvertToSql(self.select[k]) for k in ordered_distinct_vars
+            ql.ConvertToSqlForGroupBy(self.select[k])
+            for k in ordered_distinct_vars
           )
         else:
           assert False, 'Broken dialect %s, group by spec: %s' % (
@@ -471,7 +583,7 @@ def ExtractPredicateStructure(c, s):
 
   if predicate in (
       '<=', '<', '>', '>=', '!=', '&&', '||', '!', 'IsNull', 'Like',
-      'Constraint'):
+      'Constraint', 'is', 'is not', '~'):
     s.constraints.append({'call': c})
     return
 
@@ -486,10 +598,12 @@ def ExtractPredicateStructure(c, s):
     expr = field_value['value']['expression']
     var_name = s.allocator.AllocateVar('%s_%s' % (table_name, table_var))
     s.vars_map[table_name, table_var] = var_name
+    s.vars_heritage_map[table_name, table_var] = expr.get('expression_heritage')
     s.inv_vars_map[var_name] = (table_name, table_var)
     s.vars_unification.append(
         {
-            'left': {'variable': {'var_name': var_name}},
+            'left': {'variable': {'var_name': var_name},
+                     'expression_heritage': expr.get('expression_heritage')},
             'right': expr
         })
 
@@ -533,7 +647,8 @@ def ExtractInclusionStructure(inclusion, s):
               'value': {
                 'expression': {
                   'variable': {
-                    'var_name': var_name
+                    'var_name': var_name,
+                    'dont_expand': True
                   }
                 }
               }
@@ -551,7 +666,9 @@ def ExtractConjunctiveStructure(conjuncts, s):
       ExtractPredicateStructure(c['predicate'], s)
     elif 'unification' in c:
       if ('variable' in c['unification']['right_hand_side'] or
-          'variable' in c['unification']['left_hand_side']):
+          'variable' in c['unification']['left_hand_side'] or
+          'record' in c['unification']['left_hand_side'] or
+          'record' in c['unification']['right_hand_side']):
         s.vars_unification.append({
             'left': c['unification']['left_hand_side'],
             'right': c['unification']['right_hand_side']})
@@ -628,7 +745,7 @@ def InlinePredicateValuesRecursively(r, names_allocator, conjuncts):
         'Got: %s' % str(r))
 
   for k in member_index:
-    if k != 'combine':
+    if k != 'combine' and k != 'type':
       if isinstance(r[k], dict) or isinstance(r[k], list):
         InlinePredicateValuesRecursively(r[k], names_allocator, conjuncts)
 
@@ -640,7 +757,9 @@ def InlinePredicateValuesRecursively(r, names_allocator, conjuncts):
         r_predicate['predicate'] = copy.deepcopy(r['call'])
         r_predicate['predicate']['record']['field_value'].append({
             'field': 'logica_value',
-            'value': {'expression': {'variable': {'var_name': aux_var}}}
+            'value': {'expression': {
+              'variable': {'var_name': aux_var},
+              'expression_heritage': r['expression_heritage']}}
         })
         del r['call']
         r['variable'] = {'var_name': aux_var}
