@@ -43,17 +43,27 @@ import sys
 if __name__ == '__main__' and not __package__:
   from common import color
   from common import sqlite3_logica
+  from common import clingo_logica
+  from common import duckdb_logica
   from compiler import functors
   from compiler import rule_translate
   from compiler import universe
   from parser_py import parse
+  from type_inference.research import infer
+  from type_inference import type_retrieval_service_discovery
+  from tools import proposition_repl
 else:
   from .common import color
   from .common import sqlite3_logica
+  from .common import clingo_logica
+  from .common import duckdb_logica
   from .compiler import functors
   from .compiler import rule_translate
   from .compiler import universe
   from .parser_py import parse
+  from .type_inference.research import infer
+  from .type_inference import type_retrieval_service_discovery
+  from .tools import proposition_repl
 
 
 def ReadUserFlags(rules, argv):
@@ -130,13 +140,15 @@ def main(argv):
           'GoodIdea(snack: "carrots")\'')
     return 1
 
-  if len(argv) == 3 and argv[2] == 'parse':
+  if len(argv) == 3 and argv[2] in ['parse', 'infer_types', 'show_signatures',
+                                    'propositional_playground']:
     pass  # compile needs just 2 actual arguments.
   else:
     if len(argv) < 4:
       print('Not enough arguments. Run \'logica help\' for help.',
             file=sys.stderr)
       return 1
+    predicates = argv[3]
 
   if argv[1] == '-':
     filename = '/dev/stdin'
@@ -145,7 +157,9 @@ def main(argv):
 
   command = argv[2]
 
-  commands = ['parse', 'print', 'run', 'run_to_csv']
+  commands = ['parse', 'print', 'run', 'run_to_csv', 'run_in_terminal',
+              'infer_types', 'show_signatures', 'build_schema',
+              'propositional_playground', 'print_clingo', 'run_clingo']
 
   if command not in commands:
     print(color.Format('Unknown command {warning}{command}{end}. '
@@ -155,6 +169,17 @@ def main(argv):
   if not os.path.exists(filename):
     print('File not found: %s' % filename, file=sys.stderr)
     return 1
+
+  # This has to be before reading program.
+  if command == 'run_in_terminal':
+    if __name__ == '__main__' and not __package__:
+      from tools import run_in_terminal
+    else:
+      from .tools import run_in_terminal
+    artistic_table = run_in_terminal.Run(filename, predicates)
+    print(artistic_table)
+    return
+
   program_text = open(filename).read()
 
   try:
@@ -164,16 +189,60 @@ def main(argv):
     parsing_exception.ShowMessage()
     sys.exit(1)
 
+  if command == 'propositional_playground':
+    proposition_repl.Repl(program_text)
+    return
+
   if command == 'parse':
-    # No indentation to avoid file size inflation.
-    print(json.dumps(parsed_rules, sort_keys=True, indent=''))
+    # Minimal indentation for better readability of deep objects.
+    print(json.dumps(parsed_rules, sort_keys=True, indent=' '))
     return 0
 
-  predicates = argv[3]
+  if command == 'infer_types':
+    # This disallows getting types of program with type errors.
+    # logic_program = universe.LogicaProgram(parsed_rules)
+    # TODO: Find a way to get engine from program. But it should not matter
+    # for inference. It only patters for compiling.
+    typing_engine = infer.TypesInferenceEngine(parsed_rules, "psql")
+    typing_engine.InferTypes()
+    # print(parsed_rules)
+    print(json.dumps(parsed_rules, sort_keys=True, indent=' '))
+    return 0
+
+  if command == 'show_signatures':
+    try:
+      logic_program = universe.LogicaProgram(parsed_rules)
+      if not logic_program.typing_engine:
+        logic_program.RunTypechecker()
+    except infer.TypeErrorCaughtException as type_error_exception:
+      print(logic_program.typing_engine.ShowPredicateTypes())
+      type_error_exception.ShowMessage()
+      return 1
+    print(logic_program.typing_engine.ShowPredicateTypes())
+    return 0
+
+  predicates_list = predicates.split(',')
 
   user_flags = ReadUserFlags(parsed_rules, argv[4:])
 
-  predicates_list = predicates.split(',')
+  if command == 'build_schema':
+    logic_program = universe.LogicaProgram(parsed_rules, user_flags=user_flags)
+    engine = logic_program.annotations.Engine()
+    type_retrieval_service = type_retrieval_service_discovery\
+      .get_type_retrieval_service(engine, parsed_rules, predicates_list)
+    type_retrieval_service.RetrieveTypes(filename)
+    return 0
+
+  if command == 'print_clingo':
+    print(clingo_logica.Klingon(parsed_rules, predicates_list))
+    return 0
+
+  if command == 'run_clingo':
+    print(clingo_logica.RenderKlingon(
+        clingo_logica.RunClingo(
+            clingo_logica.Klingon(parsed_rules, predicates_list))))
+    return 0
+
   for predicate in predicates_list:
     try:
       logic_program = universe.LogicaProgram(
@@ -188,6 +257,12 @@ def main(argv):
     except functors.FunctorError as functor_exception:
       functor_exception.ShowMessage()
       sys.exit(1)
+    except infer.TypeErrorCaughtException as type_error_exception:
+      type_error_exception.ShowMessage()
+      sys.exit(1)
+    except parse.ParsingException as parsing_exception:
+      parsing_exception.ShowMessage()
+      sys.exit(1)
 
     if command == 'print':
       print(formatted_sql)
@@ -210,13 +285,34 @@ def main(argv):
           [preamble] + defines_and_exports + [main_predicate_sql])
         o = sqlite3_logica.RunSqlScript(statements_to_execute,
                                         format).encode()
+      elif engine == 'duckdb':
+        connection = duckdb_logica.GetConnection(logic_program)
+        cur = connection.sql(formatted_sql)
+        formatter = (sqlite3_logica.ArtisticTable
+                     if command == 'run'
+                     else sqlite3_logica.Csv)
+        o = formatter(cur.columns,
+                      cur.fetchall()).encode()
       elif engine == 'psql':
-        p = subprocess.Popen(['psql', '--quiet'] +
-                             (['--csv'] if command == 'run_to_csv' else []),
-                             stdin=subprocess.PIPE, stdout=subprocess.PIPE)
-        commands = []
-        o, _ = p.communicate(
-            '\n'.join(commands + [formatted_sql]).encode())
+        connection_str = os.environ.get('LOGICA_PSQL_CONNECTION')
+        if connection_str:
+          connection_str = os.environ.get('LOGICA_PSQL_CONNECTION')
+          import psycopg2
+          from common import psql_logica
+          connection = psycopg2.connect(connection_str)
+          cursor = psql_logica.PostgresExecute(formatted_sql, connection)
+          rows = [list(map(psql_logica.DigestPsqlType, row))
+              
+                  for row in cursor.fetchall()]
+          o = sqlite3_logica.ArtisticTable([d[0] for d in cursor.description],
+                                           rows).encode()
+        else:
+          p = subprocess.Popen(['psql', '--quiet'] +
+                              (['--csv'] if command == 'run_to_csv' else []),
+                              stdin=subprocess.PIPE, stdout=subprocess.PIPE)
+          commands = []
+          o, _ = p.communicate(
+              '\n'.join(commands + [formatted_sql]).encode())
       elif engine == 'trino':
         a = logic_program.annotations.annotations['@Engine']['trino']
         params = GetTrinoParameters(a)
